/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: XenHost.java 393 2011-10-03 13:38:05Z dangtran $
 *
 */

package org.ow2.sirocco.vmm.agent.driver.xenapi;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import javax.management.MBeanNotificationInfo;
import javax.management.Notification;
import javax.management.NotificationEmitter;
import javax.management.ObjectName;

import org.apache.log4j.Logger;
import org.ow2.sirocco.vmm.agent.domain.AbstractHost;
import org.ow2.sirocco.vmm.agent.driver.util.RemoteExec;
import org.ow2.sirocco.vmm.agent.driver.xenapi.XenVMImageStore.XenVirtualMachineImage;
import org.ow2.sirocco.vmm.agent.jmx.MBeanObjectNamer;
import org.ow2.sirocco.vmm.agent.main.AgentCommon;
import org.ow2.sirocco.vmm.api.CustomizationSpec;
import org.ow2.sirocco.vmm.api.DiskOperation;
import org.ow2.sirocco.vmm.api.InsufficientResourcesException;
import org.ow2.sirocco.vmm.api.InvalidVMConfigException;
import org.ow2.sirocco.vmm.api.NotificationTypes;
import org.ow2.sirocco.vmm.api.ResourcePartitionMXBean;
import org.ow2.sirocco.vmm.api.ResourceUsage;
import org.ow2.sirocco.vmm.api.ServerPoolMXBean;
import org.ow2.sirocco.vmm.api.VMConfigSpec;
import org.ow2.sirocco.vmm.api.VMCustomizationSpec;
import org.ow2.sirocco.vmm.api.VMMException;
import org.ow2.sirocco.vmm.api.VNICSpec;
import org.ow2.sirocco.vmm.api.VNICSpec.MacAddressAssignement;
import org.ow2.sirocco.vmm.api.VirtualCdromSpec;
import org.ow2.sirocco.vmm.api.VirtualDiskSpec;
import org.ow2.sirocco.vmm.api.VirtualMachineConfigSpec;
import org.ow2.sirocco.vmm.api.VirtualMachineImageStoreMXBean;
import org.ow2.sirocco.vmm.api.VirtualMachineMXBean;
import org.ow2.sirocco.vmm.api.VnicIPSettings;
import org.ow2.sirocco.vmm.api.VnicIPSettings.IpAssignmentMode;
import org.ow2.sirocco.vmm.api.Volume;

import com.xensource.xenapi.Connection;
import com.xensource.xenapi.Host;
import com.xensource.xenapi.HostCpu;
import com.xensource.xenapi.Network;
import com.xensource.xenapi.SR;
import com.xensource.xenapi.Types.OnCrashBehaviour;
import com.xensource.xenapi.Types.OnNormalExit;
import com.xensource.xenapi.Types.VbdMode;
import com.xensource.xenapi.Types.VbdType;
import com.xensource.xenapi.Types.VmPowerState;
import com.xensource.xenapi.VBD;
import com.xensource.xenapi.VDI;
import com.xensource.xenapi.VIF;
import com.xensource.xenapi.VM;

/**
 * XenAPI driver Host MXBean implementation
 */
public class XenHost extends AbstractHost implements NotificationEmitter {
    static Logger logger = Logger.getLogger(XenHost.class);

    // the thread pool size determines the maximum number of
    // simultaneous VM creation on the same host
    private static final int MAX_PARALELL_VM_CREATION_PER_HOST = 3;

    // one extra thread for the event reader
    private static final int THREADPOOL_SIZE = 1 + XenHost.MAX_PARALELL_VM_CREATION_PER_HOST;

    private static ExecutorService executorService = Executors.newFixedThreadPool(XenHost.THREADPOOL_SIZE);

    private List<XenVirtualMachine> vmList = new CopyOnWriteArrayList<XenVirtualMachine>();

    XenServerPool serverPool;

    private Connection connection;

    com.xensource.xenapi.Host host;

    private SR storageRepository;

    private String hostName;

    private HostPowerState hostPowerState = HostPowerState.UNKNOWN;

    private int numCPU = -1;

    private long cpuFrequencyMhz = -1;

    private String uuid;

    private RemoteExec.SshAuthInfo hostAuthInfo;

    private Map<String, String> hostAttributes;

    public static XenHost newHost(final XenServerPool resourcePool, final ObjectName objectName, final String hostName,
        final String poolUser, final String poolPassword, final int port, final Connection masterCon,
        final RemoteExec.SshAuthInfo hostAuthInfo, final Map<String, String> attributes) {
        String user = attributes.get("user");
        if (user == null) {
            user = poolUser;
            attributes.put("user", poolUser);
        }
        if (user == null) {
            throw new IllegalArgumentException("Host: " + hostName + " Missing parameter: user");
        }
        String password = attributes.get("password");
        if (password == null) {
            password = poolPassword;
            attributes.put("password", poolPassword);
        }
        if (password == null) {
            throw new IllegalArgumentException("Host: " + hostName + " Missing parameter: password");
        }

        try {
            Set<Host> hosts = Host.getAll(masterCon);
            XenHost xenHost = null;
            for (Host h : hosts) {
                if (h.getNameLabel(masterCon).equals(hostName) || h.getAddress(masterCon).equals(hostName)) {
                    xenHost = new XenHost(resourcePool, masterCon, objectName, hostName, h, hostAuthInfo, attributes);
                }
            }
            return xenHost;
        } catch (Exception ex) {
            XenHost.logger.error("Failed to establish Xen-API connection with host " + hostName + " with login " + user, ex);
            return null;
        }

    }

    private XenHost(final XenServerPool serverPool, final Connection connection, final ObjectName objectName,
        final String hostName, final Host host, final RemoteExec.SshAuthInfo hostAuthInfo,
        final Map<String, String> hostAttributes) {
        super(objectName);
        this.serverPool = serverPool;
        this.connection = connection;
        this.hostName = hostName;
        this.host = host;
        this.hostAuthInfo = hostAuthInfo;
        this.hostAttributes = new HashMap<String, String>(hostAttributes);
        try {
            this.uuid = host.getUuid(this.connection);
            this.populate();
        } catch (Exception ex) {
            XenHost.logger.debug("Host " + hostName, ex);
        }
        this.storageRepository = this.getStorageRepository();
        this.hostPowerState = HostPowerState.RUNNING;
    }

    public void removeVMFromInventory(final XenVirtualMachine vm) {
        try {
            AgentCommon.getMBeanServer().unregisterMBean(vm.getObjectName());
        } catch (Exception ex) {
            XenHost.logger.error("Cannot unregister VM MBean", ex);
        }
        this.vmList.remove(vm.getNameLabel());
        this.emitNotification(NotificationTypes.VM_INVENTORY_DEL, "Deleted", vm.getObjectName());
        XenHost.logger.info("deleted VM " + vm.getObjectName());
    }

    public boolean isConnectionLost() {
        synchronized (this.connection) {
            try {
                this.host.getAPIVersionMajor(this.connection);
            } catch (Exception ex) {
                return true;
            }
            return false;
        }
    }

    Connection getXenAPIConnection() {
        return this.connection;
    }

    private SR getStorageRepository() {
        if (this.storageRepository == null) {
            String sharedStorageRepository = this.serverPool.getSharedStorageRepository();
            try {
                Set<SR> srSet = SR.getByNameLabel(this.connection, sharedStorageRepository);
                if (srSet.isEmpty()) {
                    return null;
                }
                this.storageRepository = srSet.iterator().next();
            } catch (Exception ex) {
                XenHost.logger.error("Failed to get SR " + sharedStorageRepository, ex);
            }
        }
        return this.storageRepository;
    }

    List<XenVirtualMachine> getVMs() {
        return this.vmList;
    }

    XenVirtualMachine lookUpByUuid(final String uuid) {
        for (XenVirtualMachine vm : this.vmList) {
            if (vm.getUuid().equals(uuid)) {
                return vm;
            }
        }
        return null;
    }

    protected synchronized void addVM(final XenVirtualMachine vm) {
        this.vmList.add(vm);
    }

    protected synchronized void removeVM(final XenVirtualMachine vm) {
        this.vmList.remove(vm);
    }

    public ServerPoolMXBean getServerPool() {
        return this.serverPool;
    }

    public void postMigrateVM(final XenVirtualMachine vm, final XenHost newHost) {
        this.removeVM(vm);
        newHost.addVM(vm);
    }

    @Override
    public ObjectName getObjectName() {
        return this.objectName;
    }

    HashMap<String, String> hypervisorInfo;

    // TODO
    public Map<String, String> getHypervisorInfo() throws VMMException {
        if (this.hypervisorInfo == null) {
            this.hypervisorInfo = new HashMap<String, String>();
            try {
                Map<String, String> softwareVersions = this.host.getSoftwareVersion(this.connection);
                String xenVersion = softwareVersions.get("xen");
                if (xenVersion != null) {
                    this.hypervisorInfo.put("version", xenVersion);
                }
            } catch (Exception ex) {
            }
            this.hypervisorInfo.put("vendor", "");
            this.hypervisorInfo.put("name", "xen");
        }
        return this.hypervisorInfo;
    }

    HashMap<String, String> cpuInfo;

    public Map<String, String> getCPUInfo() throws VMMException {
        if (this.cpuInfo == null) {
            this.cpuInfo = new HashMap<String, String>();
            synchronized (this.connection) {
                try {
                    HostCpu[] cpus = this.host.getHostCPUs(this.connection).toArray(new HostCpu[0]);
                    this.numCPU = cpus.length;
                    HostCpu.Record rec = cpus[0].getRecord(this.connection);
                    this.cpuInfo.put("model", rec.modelname);
                    this.cpuFrequencyMhz = rec.speed;
                    this.cpuInfo.put("speedMHz", Long.toString(rec.speed));
                    this.cpuInfo.put("vendor", rec.vendor);
                    this.cpuInfo.put("features", rec.features);
                    this.cpuInfo.put("flags", rec.flags);
                } catch (Exception ex) {
                    throw XenDriver.translateXenAPIException(ex);
                }
            }
        }
        return this.cpuInfo;
    }

    public long getCPUFrequencyMhz() throws VMMException {
        if (this.cpuFrequencyMhz == -1) {
            this.getCPUInfo();
        }
        return this.cpuFrequencyMhz;

    }

    public int getNumCPU() throws VMMException {
        if (this.numCPU == -1) {
            this.getCPUInfo();
        }
        return this.numCPU;
    }

    public String getHostName() {
        return this.hostName;
    }

    public String getUuid() {
        return this.uuid;
    }

    public VirtualMachineImageStoreMXBean getVMImageStore() {
        return this.serverPool.getVMImageStore();
    }

    public long getFreeMemoryMB() throws VMMException {
        synchronized (this.connection) {
            try {
                return this.host.getMetrics(this.connection).getMemoryFree(this.connection) / (1024 * 1024);
            } catch (Exception ex) {
                throw XenDriver.translateXenAPIException(ex);
            }
        }
    }

    // only running VMs are attached to their host
    // halted VMs are handled in the post-create hook of the server pool
    private void populate() {
        try {
            synchronized (this.connection) {
                Set<VM> vms = VM.getAll(this.connection);
                for (VM vm : vms) {
                    VM.Record vmRecord = vm.getRecord(this.connection);
                    if (vmRecord.isATemplate || vmRecord.isControlDomain || vmRecord.powerState != VmPowerState.RUNNING) {
                        continue;
                    }
                    ObjectName name = MBeanObjectNamer.makeVirtualMachineName(
                        this.serverPool.getPath() + "/" + vm.getNameLabel(this.connection), vm.getUuid(this.connection));
                    if (AgentCommon.getMBeanServer().isRegistered(name)) {
                        continue;
                    }
                    if (!this.uuid.equals(vmRecord.residentOn.getUuid(this.connection))) {
                        continue;
                    }

                    XenVirtualMachine xenVM = new XenVirtualMachine(name, this, this.connection, vm, null);
                    xenVM.setResourcePartition(this.serverPool);
                    this.addVM(xenVM);
                    AgentCommon.getMBeanServer().registerMBean(xenVM, name);
                    XenHost.logger.info("Added XenVirtualMachineMBean " + name);
                }
            }
        } catch (Exception ex) {
            XenHost.logger.debug("Host " + this.hostName, ex);
        }
    }

    public List<VirtualMachineMXBean> getResidentVMs() {
        List<VirtualMachineMXBean> result = new ArrayList<VirtualMachineMXBean>(this.vmList);
        return result;
    }

    @Override
    public VirtualMachineMXBean createVM(final VMConfigSpec vmSpec, final boolean sync) throws InsufficientResourcesException,
        InvalidVMConfigException, VMMException {
        return this.createVM(vmSpec, this.serverPool, sync);
    }

    public VirtualMachineMXBean createVM(final VMConfigSpec vmSpecIn, final ResourcePartitionMXBean partition,
        final boolean sync) throws InvalidVMConfigException, VMMException {
        final VMConfigSpec vmSpec = new VMConfigSpec(vmSpecIn);
        XenHost.logger.info("Creating new VM [name=" + vmSpec.getName() + ",memorySizeMB=" + vmSpec.getMemorySizeMB()
            + ",diskSize=" + vmSpec.getDiskSizeMB() + ",numVCPU=" + vmSpec.getNumVCPU() + "]");

        final XenVirtualMachineImage template = (XenVirtualMachineImage) this.serverPool.getVMImageStore().lookUpByUUID(
            vmSpec.getVmImageUUID());
        if (template == null) {
            throw new InvalidVMConfigException("Invalid VM Image UUID");
        }
        final VM templateVM;
        try {
            templateVM = VM.getByUuid(this.connection, template.getUUID());
        } catch (Exception ex) {
            throw XenDriver.translateXenAPIException(ex);
        }
        if (templateVM == null) {
            throw new InvalidVMConfigException("Invalid VM Image UUID");
        }

        final String vmName = vmSpec.getName();

        if (!sync) {
            XenHost.executorService.execute(new Runnable() {
                public void run() {
                    try {
                        XenHost.this.cloneVM(vmSpec, vmName, templateVM, partition);
                    } catch (Exception ex) {
                    }
                }
            });
            return null;
        } else {
            return this.cloneVM(vmSpec, vmName, templateVM, partition);
        }
    }

    @Override
    public VirtualMachineMXBean createVirtualMachine(final VirtualMachineConfigSpec vmConfigSpec,
        final CustomizationSpec customizationSpec, final boolean startVM, final boolean sync)
        throws InsufficientResourcesException, InvalidVMConfigException, VMMException {

        this.checkVMConfigSpec(vmConfigSpec);

        VM newVM = null;
        VM.Record vmRecord = new VM.Record();

        vmRecord.actionsAfterCrash = OnCrashBehaviour.DESTROY;
        vmRecord.actionsAfterReboot = OnNormalExit.RESTART;
        vmRecord.actionsAfterShutdown = OnNormalExit.DESTROY;

        vmRecord.affinity = this.host;
        vmRecord.memoryDynamicMin = vmConfigSpec.getMemoryMB() * 1024 * 1024;
        vmRecord.memoryDynamicMax = vmConfigSpec.getMemoryMB() * 1024 * 1024;
        vmRecord.memoryStaticMin = 0L;
        vmRecord.memoryStaticMax = vmConfigSpec.getMemoryMB() * 1024 * 1024;
        vmRecord.memoryTarget = vmConfigSpec.getMemoryMB() * 1024 * 1024;

        vmRecord.nameLabel = vmConfigSpec.getName();
        vmRecord.nameDescription = "";

        vmRecord.VCPUsAtStartup = (long) vmConfigSpec.getNumVCPUs();
        vmRecord.VCPUsMax = (long) vmConfigSpec.getNumVCPUs();
        vmRecord.VCPUsParams = new HashMap<String, String>();

        vmRecord.PVBootloader = "pygrub";
        vmRecord.PVArgs = "console=hvc0 xencons=hvc";
        // vmRecord.HVMBootPolicy = "BIOS order";
        // vmRecord.HVMBootParams = new HashMap<String, String>();
        // vmRecord.HVMBootParams.put("order", "dc"); // cdrom, disk
        vmRecord.HVMShadowMultiplier = 1.0d;

        vmRecord.platform = new HashMap<String, String>();
        vmRecord.platform.put("acpi", "true");
        vmRecord.platform.put("pae", "true");
        vmRecord.platform.put("apic", "true");
        vmRecord.platform.put("viridian", "true");
        vmRecord.platform.put("timeoffset", "0");
        vmRecord.platform.put("nx", "false");

        String bootDevice = vmConfigSpec.getProperties().get("bootDevice");
        boolean bootFromCdrom = false;
        if (bootDevice != null && bootDevice.equals("cdrom")) {
            bootFromCdrom = true;
            vmRecord.PVBootloader = "eliloader";
            vmRecord.otherConfig = new HashMap<String, String>();
            vmRecord.otherConfig.put("install-repository", "cdrom");
            // XXX remove these hardcoded values
            vmRecord.otherConfig.put("install-arch", "amd64"); // or i386
            vmRecord.otherConfig.put("install-distro", "debianlike");
        }

        try {
            newVM = VM.create(this.connection, vmRecord);
        } catch (Exception ex) {
            throw XenDriver.translateXenAPIException(ex);
        }

        int userDeviceIndex = 0;

        boolean isRootDisk = true; // first disk is root disk
        for (VirtualDiskSpec diskSpec : vmConfigSpec.getDiskSpecs()) {
            Volume vol;
            XenStoragePool storagePool = (XenStoragePool) this.serverPool.getStoragePoolByName(diskSpec.getStoragePool()
                .getName());
            if (storagePool == null) {
                XenHost.logger.error("Cannot find Xen storage pool " + diskSpec.getStoragePool().getName());
                continue;
            }
            String volName = "vol-" + vmConfigSpec.getName() + "-" + System.nanoTime();
            if (diskSpec.getDiskOp() == DiskOperation.NEW) {
                vol = storagePool.createVolume(volName, diskSpec.getCapacityMB(), XenStoragePool.VHD_FORMAT);
            } else if (diskSpec.getDiskOp() == DiskOperation.CREATE_FROM) {
                vol = storagePool.createVolumeFrom(volName, diskSpec.getCapacityMB(), XenStoragePool.VHD_FORMAT,
                    diskSpec.getVolume(), true);
                if (!bootFromCdrom && customizationSpec != null) {
                    try {
                        this.customizeRootDisk(vol, customizationSpec);
                    } catch (Exception e) {
                        XenHost.logger.error("Failed to customize disk " + vol.getPath(), e);
                    }
                }
            } else {
                vol = diskSpec.getVolume();
            }

            VDI vdi = null;

            try {
                vdi = VDI.getByUuid(this.connection, vol.getKey());
            } catch (Exception ex) {
                XenHost.logger.error("Cannot locate VDI with key " + vol.getKey(), ex);
                throw XenDriver.translateXenAPIException(ex);
            }

            VBD.Record vbdRecord = new VBD.Record();
            vbdRecord.VM = newVM;
            vbdRecord.VDI = vdi;
            vbdRecord.bootable = !bootFromCdrom && userDeviceIndex == 0;
            vbdRecord.mode = VbdMode.RW;
            vbdRecord.type = VbdType.DISK;
            vbdRecord.unpluggable = !isRootDisk;
            vbdRecord.userdevice = Integer.toString(userDeviceIndex);
            vbdRecord.empty = false;

            VBD vbd = null;

            try {
                vbd = VBD.create(this.connection, vbdRecord);
            } catch (Exception ex) {
                XenHost.logger.error("Cannot create VBD", ex);
                throw XenDriver.translateXenAPIException(ex);
            }
            userDeviceIndex++;
            // skip device id 2 (reserved) and 3 (cdrom only)
            if (userDeviceIndex == 2) {
                userDeviceIndex = 4;
            }
            isRootDisk = false;
        }

        if (userDeviceIndex < 3) {
            userDeviceIndex = 3;
        }
        int firstCdromDeviceNumber = userDeviceIndex;
        if (vmConfigSpec.getCdromSpecs() != null) {
            for (VirtualCdromSpec cdromSpec : vmConfigSpec.getCdromSpecs()) {
                VDI vdi = null;
                if (cdromSpec.getIsoVolume() != null) {
                    try {
                        vdi = VDI.getByUuid(this.connection, cdromSpec.getIsoVolume().getKey());
                    } catch (Exception ex) {
                        XenHost.logger.error("Cannot locate ISO VDI with key " + cdromSpec.getIsoVolume().getKey(), ex);
                        throw XenDriver.translateXenAPIException(ex);
                    }
                }
                VBD.Record vbdRecord = new VBD.Record();
                vbdRecord.VM = newVM;
                vbdRecord.VDI = vdi;
                vbdRecord.empty = vdi == null;
                vbdRecord.bootable = bootFromCdrom && userDeviceIndex == firstCdromDeviceNumber;
                vbdRecord.mode = VbdMode.RO;
                vbdRecord.type = VbdType.CD;
                vbdRecord.unpluggable = true;
                vbdRecord.userdevice = Integer.toString(userDeviceIndex++);

                VBD vbd = null;

                try {
                    vbd = VBD.create(this.connection, vbdRecord);
                } catch (Exception ex) {
                    XenHost.logger.error("Cannot create VBD", ex);
                    throw XenDriver.translateXenAPIException(ex);
                }
            }
        }

        if (vmConfigSpec.getVnicSpecs() != null) {
            int vifDeviceNumber = 0;
            for (VNICSpec vnicSpec : vmConfigSpec.getVnicSpecs()) {
                VIF.Record vifRec = new VIF.Record();
                vifRec.VM = newVM;
                vifRec.device = Integer.toString(vifDeviceNumber++);
                if (vnicSpec.getAddressType() == MacAddressAssignement.MANUAL) {
                    vifRec.MAC = vnicSpec.getMacAddress();
                } else {
                    vifRec.MACAutogenerated = true;
                }
                try {
                    if (vnicSpec.getNetworkName() == null || vnicSpec.getNetworkName().equals("default")) {
                        vifRec.network = this.serverPool.getDefaultNetwork();
                    } else {
                        Set<Network> nets = Network.getByNameLabel(this.connection, vnicSpec.getNetworkName());
                        if (nets.size() == 0) {
                            throw new VMMException("Cannot find network " + vnicSpec.getNetworkName());
                        }
                        vifRec.network = nets.iterator().next();
                    }
                    VIF.create(this.connection, vifRec);
                } catch (Exception ex) {
                    XenHost.logger.error("Failed to create VIF", ex);
                    throw XenDriver.translateXenAPIException(ex);
                }
            }
        }

        XenVirtualMachine xenVM = null;
        synchronized (this.serverPool.eventCollector) {
            ObjectName name = MBeanObjectNamer.makeVirtualMachineName(
                XenHost.this.serverPool.getPath() + "/" + vmConfigSpec.getName(), this.uuid);
            if (!AgentCommon.getMBeanServer().isRegistered(name)) {
                xenVM = new XenVirtualMachine(name, XenHost.this, XenHost.this.connection, newVM, null);
                xenVM.setResourcePartition(this.serverPool);
                XenHost.this.addVM(xenVM);
                try {
                    AgentCommon.getMBeanServer().registerMBean(xenVM, name);
                } catch (Exception ex) {
                    XenHost.logger.error("Failed to register MBean", ex);
                }
                XenHost.this.emitNotification(NotificationTypes.VM_ADD, "Created", name);
            }
        }

        if (startVM) {
            xenVM.start();
        }
        return xenVM;
    }

    private VM controlDomain;

    private VM getDom0() throws Exception {
        if (this.controlDomain == null) {
            Set<VM> vms = VM.getByNameLabel(this.connection, "Control domain on host: " + this.hostName);
            if (vms.size() > 0) {
                this.controlDomain = vms.iterator().next();
            } else {
                XenHost.logger.error("Cannot find control domain of host " + this.hostName);
            }

            // Set<VM> vms = this.host.getResidentVMs(this.connection);
            // while (vms.iterator().hasNext()) {
            // VM vm = vms.iterator().next();
            // if (vm.getIsControlDomain(this.connection)) {
            // this.controlDomain = vm;
            // return vm;
            // }
            // }
        }
        return this.controlDomain;
    }

    private synchronized void customizeRootDisk(final Volume vol, final CustomizationSpec customizationSpec) throws Exception {
        VDI vdi = VDI.getByUuid(this.connection, vol.getKey());
        // add VBD to control domain
        VBD.Record vbdRecord = new VBD.Record();
        vbdRecord.VM = this.getDom0();
        vbdRecord.VDI = vdi;
        vbdRecord.bootable = false;
        vbdRecord.mode = VbdMode.RW;
        vbdRecord.type = VbdType.DISK;
        vbdRecord.unpluggable = true;
        vbdRecord.userdevice = "1";
        vbdRecord.empty = false;

        VBD vbd = VBD.create(this.connection, vbdRecord);
        vbd.plug(this.connection);
        String device = vbd.getDevice(this.connection) + "1";

        // launch SSH script
        RemoteExec.Result result;
        String script = AgentCommon.getScript(XenDriver.NAME, XenDriver.CUSTOMIZE_DISK_SCRIPT);
        StringBuilder command = new StringBuilder("bash -s ");
        command.append("device " + device + " ");
        if (customizationSpec.getGuestOsHostName() != null && !customizationSpec.getGuestOsHostName().equals("")) {
            command.append(" hostname " + customizationSpec.getGuestOsHostName());
        }
        if (customizationSpec.getVnicIpSettingsList() != null && customizationSpec.getVnicIpSettingsList().size() > 0) {
            VnicIPSettings nic0Settings = customizationSpec.getVnicIpSettingsList().get(0);
            if (nic0Settings.getIpAssignmentMode() == IpAssignmentMode.DHCP) {
                command.append("  net eth0/dhcp");
            } else if (nic0Settings.getIpAssignmentMode() == IpAssignmentMode.FIXED) {
                command.append("  net eth0/static/" + nic0Settings.getIpAddress() + "/" + nic0Settings.getSubnetMask() + "/"
                    + nic0Settings.getGateway());
            }
        }
        if (customizationSpec.getGuestOsCustomizationParams() != null) {
            for (String param : customizationSpec.getGuestOsCustomizationParams().keySet()) {
                if (param.equals("sshkey")) {
                    command.append(" sshkey \"" + customizationSpec.getGuestOsCustomizationParams().get(param) + "\"");
                } else {
                    command.append(" userdata \"" + param + "\" \""
                        + customizationSpec.getGuestOsCustomizationParams().get(param) + "\"");
                }
            }
        }
        command.append(" << 'END'\n");
        command.append(script);
        command.append("\n");
        command.append("END");
        XenHost.logger.debug("Customizing disk " + vol.getPath() + " with command \n" + command.toString());
        result = RemoteExec.commandAsRoot(this.getHostName(), this.hostAuthInfo, command.toString());
        if (result.exitCode != 0) {
            XenHost.logger.error("Failed to customize disk image: " + result.error);
            throw new Exception(result.error);
        }

        // remove vbd
        vbd.unplug(this.connection);
        vbd.destroy(this.connection);
    }

    public VirtualMachineMXBean cloneVM(final VM vm, final String clonedVmName, final VMCustomizationSpec custSpec,
        final ResourcePartitionMXBean partition, final boolean sync) throws InsufficientResourcesException, VMMException {
        if (!sync) {
            XenHost.executorService.execute(new Runnable() {
                public void run() {
                    try {
                        XenHost.this.cloneVM(null, clonedVmName, vm, XenHost.this.serverPool);
                    } catch (Exception ex) {
                    }
                }
            });
            return null;
        } else {
            return this.cloneVM(null, clonedVmName, vm, this.serverPool);
        }
    }

    private VirtualMachineMXBean cloneVM(final VMConfigSpec vmSpec, final String vmName, final VM templateVM,
        final ResourcePartitionMXBean partition) throws InvalidVMConfigException, VMMException {
        try {
            VM newVM;
            String uuid;
            // synchronized (this.connection) {
            newVM = templateVM.createClone(this.connection, vmName);

            if (vmSpec != null) {
                // network configuration
                Set<Network> networks = Network.getAll(this.connection);
                Network network = null;
                // XXX hardcoded bridge name xenbr0
                for (Network net : networks) {
                    String bridge = net.getBridge(this.connection);
                    XenHost.logger.debug("Network bridge=" + bridge);
                    if (bridge.equals("xenbr0")) {
                        network = net;
                        break;
                    }
                }
                if (network == null) {
                    throw new VMMException("Failed to find bridge xenbr0");
                }
                VIF.Record newvifrecord = new VIF.Record();
                newvifrecord.VM = newVM;
                newvifrecord.network = network;
                newvifrecord.device = "0";
                newvifrecord.MTU = 1500L;
                VIF.create(this.connection, newvifrecord);

                // set VM storage repository

                Map<String, String> otherConfig = newVM.getOtherConfig(this.connection);
                String disks = otherConfig.get("disks");
                disks = disks.replace("sr=\"\"", "sr=\"" + this.storageRepository.getUuid(this.connection) + "\"");
                otherConfig.put("disks", disks);
                newVM.setOtherConfig(this.connection, otherConfig);

                // to boot non-interactively
                newVM.setPVArgs(this.connection, "noninteractive");

                // provision VM
                newVM.provision(this.connection);
            }

            uuid = newVM.getUuid(this.connection);
            // }

            synchronized (this.serverPool.eventCollector) {
                ObjectName name = MBeanObjectNamer.makeVirtualMachineName(XenHost.this.serverPool.getPath() + "/" + vmName,
                    uuid);
                if (!AgentCommon.getMBeanServer().isRegistered(name)) {
                    XenVirtualMachine xenVM = new XenVirtualMachine(name, XenHost.this, XenHost.this.connection, newVM, null);
                    xenVM.setResourcePartition(partition);
                    XenHost.this.addVM(xenVM);
                    AgentCommon.getMBeanServer().registerMBean(xenVM, name);
                    XenHost.this.emitNotification(NotificationTypes.VM_ADD, "Created", name);
                }
                return this.lookUpByUuid(uuid);
            }
        } catch (Exception ex) {
            throw XenDriver.translateXenAPIException(ex);
        }
    }

    public long getTotalMemoryMB() throws VMMException {
        synchronized (this.connection) {
            try {
                return this.host.getMetrics(this.connection).getMemoryTotal(this.connection) / (1024 * 1024);
            } catch (Exception ex) {
                throw XenDriver.translateXenAPIException(ex);
            }
        }
    }

    public float[] getLoadPerCPU() throws VMMException {
        float[] result = new float[0];
        synchronized (this.connection) {
            try {
                Set<HostCpu> cpus = this.host.getHostCPUs(this.connection);
                result = new float[cpus.size()];
                int i = 0;
                for (HostCpu cpu : cpus) {
                    result[i++] = (float) (cpu.getUtilisation(this.connection) * 100);
                }
            } catch (Exception ex) {
                throw XenDriver.translateXenAPIException(ex);
            }
        }
        return result;
    }

    public float getCPULoad() throws VMMException {
        float aggregatedCPULoad = 0;
        for (XenVirtualMachine vm : this.vmList) {
            aggregatedCPULoad += vm.getCPULoad();
        }
        return aggregatedCPULoad;
    }

    public Map<String, Float> getVMCPULoads() throws VMMException {
        HashMap<String, Float> cpuLoads = new HashMap<String, Float>();
        for (XenVirtualMachine vm : this.vmList) {
            cpuLoads.put(vm.getNameLabel(), vm.getCPULoad());
        }
        return cpuLoads;
    }

    @Override
    public MBeanNotificationInfo[] getNotificationInfo() {
        return new MBeanNotificationInfo[] {new MBeanNotificationInfo(new String[] {NotificationTypes.VM_STATE_CHANGE,
            NotificationTypes.VM_ADD, NotificationTypes.VM_DEL, NotificationTypes.VM_INVENTORY_DEL, NotificationTypes.ERROR,
            NotificationTypes.LOG}, Notification.class.getName(), "Host event")};
    }

    public Map<String, ResourceUsage> getVMResourceUsage() throws VMMException {
        HashMap<String, ResourceUsage> result = new HashMap<String, ResourceUsage>();
        for (XenVirtualMachine vm : this.vmList) {
            result.put(vm.getNameLabel(), vm.getResourceUsage());
        }
        return result;
    }

    @Override
    public Map<String, String> getAttributes() {
        return this.hostAttributes;
    }

    @Override
    public void release() {
        this.connection.dispose();
        while (this.getResidentVMs().size() > 0) {
            this.removeVMFromInventory((XenVirtualMachine) this.getResidentVMs().get(0));
        }
    }

    /*
     * (non-Javadoc)
     * @see org.ow2.sirocco.vmm.api.HostMXBean#getHostState()
     */
    public HostPowerState getHostState() throws VMMException {
        return this.hostPowerState;
    }

    @Override
    public void start() {
        // TODO Auto-generated method stub
        this.hostPowerState = HostPowerState.RUNNING;
    }

    @Override
    public void stop() {
        // TODO Auto-generated method stub
        this.hostPowerState = HostPowerState.HALTED;
    }
}
