/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: XenStoragePool.java 156 2011-07-19 16:36:14Z dangtran $
 *
 */

package org.ow2.sirocco.vmm.agent.driver.xenapi;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.management.ObjectName;

import org.apache.log4j.Logger;
import org.ow2.sirocco.vmm.agent.domain.AbstractStoragePool;
import org.ow2.sirocco.vmm.api.InsufficientResourcesException;
import org.ow2.sirocco.vmm.api.StoragePoolMXBean;
import org.ow2.sirocco.vmm.api.VMMException;
import org.ow2.sirocco.vmm.api.Volume;

import com.xensource.xenapi.Connection;
import com.xensource.xenapi.SR;
import com.xensource.xenapi.Types;
import com.xensource.xenapi.VDI;

public class XenStoragePool extends AbstractStoragePool implements StoragePoolMXBean {
    static Logger logger = Logger.getLogger(XenStoragePool.class);

    public static final String VHD_FORMAT = "vhd";

    public static final String ISO_FORMAT = "iso";

    private SR sr;

    private SR.Record srRecord;

    private String poolName;

    private Map<String, String> properties;

    private Connection connection;

    private Map<String, Volume> volumeMap = new HashMap<String, Volume>();

    public XenStoragePool(final ObjectName objectName, final String poolName, final Connection connection,
        final Map<String, String> properties) throws Exception {
        super(objectName);
        Set<SR> storageRepos = SR.getByNameLabel(connection, poolName);
        if (storageRepos.size() == 0) {
            throw new Exception("SR " + poolName + " not found");
        }
        this.sr = storageRepos.iterator().next();
        this.srRecord = this.sr.getRecord(connection);
        this.poolName = poolName;
        this.properties = new HashMap<String, String>(properties);
        this.connection = connection;
        this.populate();
    }

    @Override
    public String getName() {
        return this.poolName;
    }

    @Override
    public String[] getSupportedDiskImageFormats() {
        if (this.srRecord.type.equals("iso")) {
            return new String[] {XenStoragePool.ISO_FORMAT};
        } else {
            return new String[] {XenStoragePool.VHD_FORMAT};
        }
    }

    @Override
    public long getCapacityMB() throws VMMException {
        try {
            return this.sr.getPhysicalSize(this.connection) / (1024 * 1024);
        } catch (Exception ex) {
            throw new VMMException(ex.getMessage());
        }
    }

    @Override
    public long getFreeSpaceMB() throws VMMException {
        try {
            return this.getCapacityMB() - this.sr.getPhysicalUtilisation(this.connection) / (1024 * 1024);
        } catch (Exception ex) {
            throw new VMMException(ex.getMessage());
        }
    }

    @Override
    public String getType() {
        return "file-backed";
    }

    private synchronized void populate() throws Exception {
        this.volumeMap.clear();
        Set<VDI> virtualDisks = this.sr.getVDIs(this.connection);
        for (VDI vdi : virtualDisks) {
            VDI.Record vdiRec = vdi.getRecord(this.connection);
            if (vdiRec.managed) {
                Volume vol = this.vdiToVolume(vdiRec);
                XenStoragePool.logger.debug("Added volume " + vol);
                this.volumeMap.put(vol.getKey(), vol);
            }
        }
    }

    private Volume vdiToVolume(final VDI.Record vdiRec) throws Exception {
        Volume vol = new Volume();
        vol.setCapacityMB(vdiRec.virtualSize / (1024 * 1024));
        vol.setAllocationMB(vdiRec.physicalUtilisation / (1024 * 1024));
        vol.setFormat(this.srRecord.type.equals("iso") ? XenStoragePool.ISO_FORMAT : XenStoragePool.VHD_FORMAT);
        vol.setKey(vdiRec.uuid);
        vol.setStoragePool(this);
        vol.setName(vdiRec.nameLabel);
        vol.setDescription(vdiRec.nameDescription);
        return vol;
    }

    @Override
    public synchronized List<Volume> getVolumes() throws VMMException {
        return new ArrayList<Volume>(this.volumeMap.values());
    }

    @Override
    public synchronized Volume getVolumeByKey(final String key) {
        return this.volumeMap.get(key);
    }

    @Override
    public void destroyVolume(final String key) throws VMMException {
        Volume vol = this.volumeMap.remove(key);
        if (vol == null) {
            throw new VMMException("Volume with key=" + key + " not found");
        }
        try {
            VDI vdi = VDI.getByUuid(this.connection, key);
            vdi.destroy(this.connection);
        } catch (Exception ex) {
            throw new VMMException("Cannot destroy volume with key " + key + ": " + ex.getMessage());
        }

    }

    @Override
    public Volume createVolume(final String name, final long capacityMB, final String format) throws VMMException {
        if (!format.equals(XenStoragePool.VHD_FORMAT)) {
            throw new VMMException("Unsupported format");
        }
        try {
            VDI.Record vdiRec = new VDI.Record();
            vdiRec.nameLabel = name;
            vdiRec.SR = this.sr;
            vdiRec.virtualSize = capacityMB * 1024 * 1024;
            vdiRec.type = Types.VdiType.USER;
            VDI vdi = VDI.create(this.connection, vdiRec);
            vdiRec = vdi.getRecord(this.connection);
            Volume volume = this.vdiToVolume(vdiRec);
            synchronized (this) {
                this.volumeMap.put(volume.getKey(), volume);
            }
            return volume;
        } catch (Exception ex) {
            XenStoragePool.logger.error("Failed to create volume", ex);
            throw new VMMException("Failed to create volume: " + ex.getMessage());
        }
    }

    // XXX copyOnWrite param is ignored (always true)
    @Override
    public Volume createVolumeFrom(final String name, final long capacityMB, final String format, final Volume parentVolume,
        final boolean copyOnWrite) throws VMMException {
        if (!format.equals(XenStoragePool.VHD_FORMAT)) {
            throw new VMMException("Unsupported format");
        }
        try {
            VDI parentVdi = VDI.getByUuid(this.connection, parentVolume.getKey());
            if (parentVdi == null) {
                throw new VMMException("Cannot find volume with key " + parentVolume.getKey());
            }
            Map<String, String> driverParams = new HashMap<String, String>();
            VDI vdi = parentVdi.createClone(this.connection, driverParams);
            vdi.setNameLabel(this.connection, name);
            VDI.Record vdiRec = vdi.getRecord(this.connection);
            Volume volume = this.vdiToVolume(vdiRec);
            synchronized (this) {
                this.volumeMap.put(volume.getKey(), volume);
            }
            return volume;
        } catch (Exception ex) {
            XenStoragePool.logger.error("Failed to create volume", ex);
            throw new VMMException("Failed to create volume: " + ex.getMessage());
        }
    }

    @Override
    public Volume uploadVolume(final String url, final String format, final String name, final String description,
        final String checkSum) throws InsufficientResourcesException, VMMException {
        // TODO Auto-generated method stub
        return null;
    }

    @Override
    public Map<String, String> getProperties() {
        return this.properties;
    }

    @Override
    public synchronized void refresh() {
        try {
            this.sr.scan(this.connection);
            this.populate();
        } catch (Exception ex) {
            XenStoragePool.logger.error("failed to refresh SR " + this.poolName, ex);
        }
    }

}
